/*
 * 2017 Petr Socha
 * Czech Technical University in Prague
 * Faculty of Information Technology
 * Department of Digital Design
 * 
 * This alfa version is intended for internal testing use only! 
 */

#include <stdio.h>
#include <stdlib.h>
#include <inttypes.h>
#include <omp.h>
#include <math.h>
#include <tclap/CmdLine.h>

typedef int16_t traceType;
typedef uint8_t powermodelType;
typedef double  floatType;

int main(int argc, char * argv[]){
    
    TCLAP::CmdLine cmd("This program is used to compute correlation coefficients between precomputed power model and measured power traces when attacking AES using correlation power analysis.", ' ', "0.5alfa");

    TCLAP::ValueArg<std::string> tracesArg("t", "traces", "Name of input binary file with measured power traces", true, "", "Name of an existing file");
    TCLAP::ValueArg<std::string> powerArg("p", "powermodel", "Name of input binary file with precomputed power model", true, "", "Name of an existing file");
    TCLAP::ValueArg<std::string> outputArg("o", "output", "Name of output binary file where correlation coefficient matrices will be saved", true, "", "Name of a file to be created");    
    TCLAP::ValueArg<int> measurementsArg("n", "numberoftraces", "Number of power traces", true, 0, "Natural number");
    TCLAP::ValueArg<int> samplesArg("s", "numberofsamples", "Number of samples per trace", true, 0, "Natural number");        
    TCLAP::ValueArg<int> bufferSizeArg("m", "maxbuffersize", "Maximum amount of RAM to be used, in MiB.", false, 0, "Natural number");
    
    cmd.add(bufferSizeArg);
    cmd.add(outputArg);
    cmd.add(powerArg);    
    cmd.add(tracesArg);
    cmd.add(samplesArg);
    cmd.add(measurementsArg);
        
    try {
        cmd.parse(argc, argv);
    }
    catch (TCLAP::ArgException &e) {
        std::cerr << "error: " << e.error() << " for arg " << e.argId() << std::endl;
    }
    
    
    
    size_t tracesN = measurementsArg.getValue(); 
    size_t samplesN = samplesArg.getValue();    
    
    if(tracesN < 0 || samplesN < 0 || bufferSizeArg.getValue() < 0){
        fprintf(stderr, "Argument error: Negative number where natural number expected.\n");
        return 1;
    }
                    
    size_t tracesNMod = tracesN; // number of traces to be loaded at once, may change in following branching
    
    // when asking for maximum buffer size, recompute tracesNMod
    if(bufferSizeArg.isSet()){ 
        size_t minMemMiB = (((sizeof(powermodelType) * 256 * tracesN) + 2*(sizeof(floatType) * samplesN) + 2*(sizeof(floatType) * 256) + (sizeof(floatType) * 256 * samplesN) + 10*(sizeof(traceType) * samplesN)) / 1024 / 1024) + 1;        // minimum amount of memory needed to store at least 10 traces and the working variables
        
        if((size_t)bufferSizeArg.getValue() < minMemMiB){ // if asking for less memory, error
            fprintf(stderr, "Argument error: Minimum size of memory buffer in this case is %ld MiB.\n", minMemMiB);
            return 2;
        }
        
        size_t maxTracesMemory = (bufferSizeArg.getValue() * 1024 * 1024) - ((minMemMiB * 1024 * 1024) - 10*(sizeof(traceType) * samplesN)); // amount of memory available to store measured traces
        tracesNMod = tracesNMod <= (maxTracesMemory / (sizeof(traceType) * samplesN)) ? tracesNMod : (maxTracesMemory / (sizeof(traceType) * samplesN)); // number of traces to be loaded at once
    }
        
    //fprintf(stderr, "Loading %d traces at once to %d.\n", tracesNMod, sizeof(traceType) * samplesN * tracesNMod);
    //fprintf(stderr, "Could be loading %d traces at once to %d.\n", tracesN, sizeof(traceType) * samplesN * tracesN);        
    
    // open files
    FILE * tracesFile = fopen(tracesArg.getValue().c_str(), "rb");
    FILE * powermodelFile = fopen(powerArg.getValue().c_str(), "rb");
    FILE * correlationFile = fopen(outputArg.getValue().c_str(), "wb");
    
    if(tracesFile == NULL || powermodelFile == NULL || correlationFile == NULL){
        fprintf(stderr, "Filesystem error: cannot open files.\n");
        return 3;
    }
    
    // alloc data memory
    traceType * traces = (traceType *) malloc(sizeof(traceType) * samplesN * tracesNMod);  // alloc right amount for defined no of traces to be loaded at once
    powermodelType * powermodel = (powermodelType *) malloc(sizeof(powermodelType) * 256 * tracesN); // powermodel for one Byte
    
    if(traces == NULL || powermodel == NULL) {
        fprintf(stderr, "Error: Not enough memory.\n");
        return 4;
    }                
            
    // alloc working variables
    floatType * tAvg = (floatType *) malloc(sizeof(floatType) * samplesN);    // vector for computation of EX
    floatType * tMSum = (floatType *) malloc(sizeof(floatType) * samplesN);   // vector for computation of varX
    
    floatType * pAvg = (floatType *) malloc(sizeof(floatType) * 256);
    floatType * pMSum = (floatType *) malloc(sizeof(floatType) * 256);
    
    floatType * CSum = (floatType *) malloc(sizeof(floatType) * 256 * samplesN);  // matrix for computation of covariance
    
    if(tAvg == NULL || tMSum == NULL || pAvg == NULL || pMSum == NULL || CSum == NULL) {
        fprintf(stderr, "Error: Not enough memory.\n");
        return 5;
    }
        
    // alloc space for keyguess
    unsigned char keyGuess[32];        
    
    double start_time = omp_get_wtime();
    
    // attacking one byte at a time: we need 1/16 of the power model and all the measured power traces, for each byte attacked
    // only (int)tracesNMod measured power traces are loaded at once (all the traces or less, when memory limit is set)
    for(size_t byte = 0; byte < 16; byte++){
        
        // attacking another byte, read new power model and clean-up global vars
        if(fread(powermodel, sizeof(powermodelType), 256 * tracesN, powermodelFile) != 256 * tracesN){
            fprintf(stderr, "Error: Not enough data (power model).\n");
            return 6;
        }        
        for(size_t i = 0; i < samplesN; i++){  
            tAvg[i] = 0;
            tMSum[i] = 0;
            // These vars dont need to be computed 16 times, however computation of CSum requires pass thru the measured traces again for every byte anyway.
            // The time penalty is unsigficicant, hence for simplicity (in this alfaversion), the computation is repeated.
        }        
        for(size_t i = 0; i < 256; i++){
            pAvg[i] = 0;
            pMSum[i] = 0;
        }        
        for(size_t i = 0; i < 256*samplesN; i++){
            CSum[i] = 0;
        }        
        
        // we go thru all the traces 16 times, for each attacked byte once        
        fseek(tracesFile, 0, 0);
        
        // load first group of traces, more will be loaded later, if needed
        if(fread(traces, sizeof(traceType), samplesN * tracesNMod, tracesFile) != samplesN * tracesNMod){
            fprintf(stderr, "Error: Not enough data (power traces).\n");
            return 7;
        }            
        
        for(size_t measure = 0; measure < tracesN; measure++){   
                        
            // bottle-neck            
            #pragma omp parallel for
            for(size_t sample = 0; sample < samplesN; sample++){                    
                            
                for(size_t key = 0; key < 256; key++){                    
                                       
                    CSum[ sample*256 + key ] += ((floatType)measure / (floatType)(measure + 1)) * ( traces[(measure % tracesNMod)*samplesN + sample] - tAvg[sample] ) * ( powermodel[measure*256 + key] - pAvg[key]);
                    
                }                                
                
            }                                                
            
            for(size_t key = 0; key < 256; key++){
                floatType temp = 0;
                temp = ((floatType)powermodel[measure*256 + key] - pAvg[key]);
                pAvg[key] += ( temp  / (floatType)(measure + 1));
                pMSum[key] += temp * ((floatType)powermodel[measure*256 + key] - pAvg[key]);
            }
            
            for(size_t sample = 0; sample < samplesN; sample++){
                floatType temp = 0;
                temp = ((floatType)traces[(measure % tracesNMod) * samplesN + sample] - tAvg[sample]);
                tAvg[sample] += ( temp / (floatType)(measure + 1));
                tMSum[sample] += temp * ((floatType)traces[(measure % tracesNMod) * samplesN + sample] - tAvg[sample]);
            }
            
                        
            // when addressing traces by measure#, modulo by number of traces loaded at once
            
            // when accessing last element of traces array (last loaded trace), load new traces[] content                    
            if((((measure + 1 ) % tracesNMod) == 0) && ((measure + 1) < tracesN)){ // we've got stuff to load
                
                size_t howMuchToLoad = tracesNMod <= (tracesN - (measure+1)) ? tracesNMod : (tracesN - (measure+1)); // load tracesNMod or all the rest

                if(fread(traces, sizeof(traceType), samplesN * howMuchToLoad, tracesFile) != samplesN * howMuchToLoad){
                    fprintf(stderr, "Error: Not enough data (power traces).\n");
                    return 8;
                }   
            }            
            
        }        
        
        // TODO: following is not neccesary, since we are only looking for max/min values of correlation coeffecients... HOWEVER, output of the algorithm is supposed to be correlation matrix, which would not be obtained without following sqrts... with time performance in mind, this could be parametricized somehow, in the future
        for(size_t sample = 0; sample < samplesN; sample++){
            tMSum[sample] = sqrt(tMSum[sample]);
        }
        for(size_t key =  0; key < 256; key++){
            pMSum[key] = sqrt(pMSum[key]);
        }
        
        // vector xAvg contains EXs, vector xMSum contains sqrt(varX)s, matrix CSum contains covariances        
        
        double max = -1;
        double min = +1;
        //int maxSample;    // TO BE used
        //int minSample;
        int maxKey;
        int minKey;
        
        // calculate correlation matrix
        for(size_t sample = 0; sample < samplesN; sample++){
            
            for(size_t key = 0; key < 256; key++){                                
                
                CSum[ sample*256 + key ] /= (tMSum[sample] * pMSum[key]);
                
                // CSum now containing Pearson correlation coefficients

                // look for keyguess
                if (CSum[ sample*256 + key ] > max) {
                    max = CSum[ sample*256 + key ];
                    //maxSample = sample;
                    maxKey = key;
                }
                else if (CSum[ sample*256 + key ] < min) {
                    min = CSum[ sample*256 + key ];
                    //minSample = sample;
                    minKey = key;
                }
                
            }
        }
        
        // save the correlation matrix
        if(fwrite(CSum, sizeof(floatType), samplesN*256, correlationFile) != samplesN*256){
            fprintf(stderr, "Filesystem error: Not enough drive space.");
            return 9;
        }
        
        keyGuess[byte] = maxKey;
        keyGuess[byte + 16] = minKey;
        
    }
    
    double time = omp_get_wtime() - start_time;
    fprintf(stderr, "Job done, %lfs elapsed.\n", time);
    
    for (int i = 0; i < 16; i++) printf("%.2x", keyGuess[i]);
    
    printf("\n");    
    
    fclose(tracesFile);
    fclose(powermodelFile);
    fflush(correlationFile);
    fclose(correlationFile);
    free(traces);
    free(powermodel);
    free(tAvg);
    free(tMSum);
    free(pAvg);
    free(pMSum);
    free(CSum);
    
    return 0;
}
